@@ -7,14 +7,11 @@
 #include "contrib/ucw/mempool.h"
 #include "daemon/engine.h"
 #include "daemon/io.h"
+#include "daemon/proxyv2.h"
 #include "daemon/network.h"
 #include "daemon/udp_queue.h"
 #include "daemon/worker.h"
 
-#ifdef ENABLE_DOH2
-#include "daemon/http.h"
-#endif
-
 #include "lib/defines.h"
 #include "lib/dnssec.h"
 #include "lib/log.h"
@@ -447,9 +444,9 @@ int main(int argc, char **argv)
 {
 	kr_log_group_reset();
 	if (setvbuf(stdout, NULL, _IONBF, 0) || setvbuf(stderr, NULL, _IONBF, 0)) {
-		kr_log_error(SYSTEM, "failed to to set output buffering (ignored): %s\n",
+		kr_log_error(SYSTEM, "failed to set output buffering (ignored): %s\n",
 				strerror(errno));
-		fflush(stderr);
+		(void)fflush(stderr);
 	}
 	if (strcmp("linux", OPERATING_SYSTEM) != 0)
 		kr_log_warning(SYSTEM, "Knot Resolver is tested on Linux, other platforms might exhibit bugs.\n"
@@ -512,7 +509,7 @@ int main(int argc, char **argv)
 	if (ret) {
 		kr_log_error(SYSTEM, "failed to get or set file-descriptor limit: %s\n",
 				strerror(errno));
-	} else if (rlim.rlim_cur < 512*1024) {
+	} else if (rlim.rlim_cur < (rlim_t)512 * 1024) {
 		kr_log_warning(SYSTEM, "warning: hard limit for number of file-descriptors is only %ld but recommended value is 524288\n",
 				(long)rlim.rlim_cur);
 	}
@@ -583,12 +580,6 @@ int main(int argc, char **argv)
 				uv_strerror(ret));
 	}
 
-	io_protolayers_init();
-	tls_protolayers_init();
-#ifdef ENABLE_DOH2
-	http_protolayers_init();
-#endif
-
 	/* Start listening, in the sense of network_listen_fd(). */
 	if (start_listening(&the_args->fds) != 0) {
 		ret = EXIT_FAILURE;
