@@ -24,10 +24,10 @@
 #include "daemon/worker.h"
 #include "daemon/session2.h"
 
-#define EPHEMERAL_CERT_EXPIRATION_SECONDS_RENEW_BEFORE (60*60*24*7)
+#define EPHEMERAL_CERT_EXPIRATION_SECONDS_RENEW_BEFORE ((time_t)60*60*24*7)
 #define GNUTLS_PIN_MIN_VERSION  0x030400
 #define UNWRAP_BUF_SIZE 131072
-#define TLS_CHUNK_SIZE (16 * 1024)
+#define TLS_CHUNK_SIZE ((size_t)16 * 1024)
 
 #define VERBOSE_MSG(cl_side, ...)\
 	if (cl_side) \
@@ -35,9 +35,9 @@
 	else \
 		kr_log_debug(TLS, __VA_ARGS__);
 
-static const gnutls_datum_t tls_grp_alpn[PROTOLAYER_GRP_COUNT] = {
-	[PROTOLAYER_GRP_DOTLS] = { (uint8_t *)"dot", 3 },
-	[PROTOLAYER_GRP_DOHTTPS] = { (uint8_t *)"h2", 2 },
+static const gnutls_datum_t tls_grp_alpn[KR_PROTO_COUNT] = {
+	[KR_PROTO_DOT] = { (uint8_t *)"dot", 3 },
+	[KR_PROTO_DOH] = { (uint8_t *)"h2", 2 },
 };
 
 typedef enum tls_client_hs_state {
@@ -71,22 +71,24 @@ static int client_verify_certificate(gnutls_session_t tls_session);
 static struct tls_credentials *tls_credentials_reserve(struct tls_credentials *tls_credentials);
 
 /**
- * Set mandatory security settings from
- * https://tools.ietf.org/html/draft-ietf-dprive-dtls-and-tls-profiles-11#section-9
+ * Set restrictions on TLS features, in particular ciphers.
+ *
+ * We explicitly disable features according to:
+ * https://datatracker.ietf.org/doc/html/rfc8310#section-9
+ * in case the gnutls+OS defaults allow them.
  * Performance optimizations are not implemented at the moment.
+ *
+ * OS defaults are taken into account, e.g. on Red Hat family there's
+ * /etc/crypto-policies/back-ends/gnutls.config and update-crypto-policies tool.
  */
 static int kres_gnutls_set_priority(gnutls_session_t session) {
-	static const char * const priorities =
-		"NORMAL:" /* GnuTLS defaults */
-		"-VERS-TLS1.0:-VERS-TLS1.1:" /* TLS 1.2 and higher */
-		 /* Some distros by default allow features that are considered
-		  * too insecure nowadays, so let's disable them explicitly. */
-		"-VERS-SSL3.0:-ARCFOUR-128:-COMP-ALL:+COMP-NULL";
+	static const char * const extra_prio =
+		"-VERS-TLS1.0:-VERS-TLS1.1:-VERS-SSL3.0:-ARCFOUR-128:-COMP-ALL:+COMP-NULL";
 	const char *errpos = NULL;
-	int err = gnutls_priority_set_direct(session, priorities, &errpos);
+	int err = gnutls_set_default_priority_append(session, extra_prio, &errpos, 0);
 	if (err != GNUTLS_E_SUCCESS) {
 		kr_log_error(TLS, "setting priority '%s' failed at character %zd (...'%s') with %s (%d)\n",
-			     priorities, errpos - priorities, errpos, gnutls_strerror_name(err), err);
+			     extra_prio, errpos - extra_prio, errpos, gnutls_strerror_name(err), err);
 	}
 	return err;
 }
@@ -216,6 +218,11 @@ static ssize_t kres_gnutls_vec_push(gnutls_transport_ptr_t h, const giovec_t * i
 		return 0;
 	}
 
+	if (kr_fails_assert(iovcnt > 0)) {
+		errno = EINVAL;
+		return -1;
+	}
+
 	size_t total_len = 0;
 	for (int i = 0; i < iovcnt; i++)
 		total_len += iov[i].iov_len;
@@ -226,9 +233,9 @@ static ssize_t kres_gnutls_vec_push(gnutls_transport_ptr_t h, const giovec_t * i
 	push_ctx->sess_data = tls;
 	memcpy(push_ctx->iov, iov, sizeof(struct iovec[iovcnt]));
 
-	session2_wrap_after(tls->h.session, PROTOLAYER_PROTOCOL_TLS,
-			protolayer_iovec(push_ctx->iov, iovcnt, true), NULL,
-			kres_gnutls_push_finished, push_ctx);
+	session2_wrap_after(tls->h.session, PROTOLAYER_TYPE_TLS,
+			protolayer_payload_iovec(push_ctx->iov, iovcnt, true),
+			NULL, kres_gnutls_push_finished, push_ctx);
 
 	return total_len;
 }
@@ -258,7 +265,7 @@ static void tls_handshake_success(struct pl_tls_sess_data *tls,
 		}
 	}
 	if (!tls->first_handshake_done) {
-		session2_event_after(session, PROTOLAYER_PROTOCOL_TLS,
+		session2_event_after(session, PROTOLAYER_TYPE_TLS,
 				PROTOLAYER_EVENT_CONNECT, NULL);
 		tls->first_handshake_done = true;
 	}
@@ -440,7 +447,7 @@ static int str_replace(char **where_ptr, const char *with)
 	return kr_ok();
 }
 
-static time_t _get_end_entity_expiration(gnutls_certificate_credentials_t creds)
+static time_t get_end_entity_expiration(gnutls_certificate_credentials_t creds)
 {
 	gnutls_datum_t data;
 	gnutls_x509_crt_t cert = NULL;
@@ -512,7 +519,7 @@ int tls_certificate_set(const char *tls_cert, const char *tls_key)
 		return kr_error(EINVAL);
 	}
 	/* record the expiration date: */
-	tls_credentials->valid_until = _get_end_entity_expiration(tls_credentials->credentials);
+	tls_credentials->valid_until = get_end_entity_expiration(tls_credentials->credentials);
 
 	/* Exchange the x509 credentials */
 	struct tls_credentials *old_credentials = the_network->tls_credentials;
@@ -782,7 +789,7 @@ static int client_verify_pin(const unsigned int cert_list_size,
  *
  * \returns GNUTLS_E_SUCCESS if certificate chain is valid, any other value is an error
  */
-static int client_verify_certchain(gnutls_session_t tls_session, const char *hostname)
+static int client_verify_certchain(struct pl_tls_sess_data *tls, const char *hostname)
 {
 	if (kr_fails_assert(hostname)) {
 		kr_log_error(TLSCLIENT, "internal config inconsistency: no hostname set\n");
@@ -790,27 +797,30 @@ static int client_verify_certchain(gnutls_session_t tls_session, const char *hos
 	}
 
 	unsigned int status;
-	int ret = gnutls_certificate_verify_peers3(tls_session, hostname, &status);
+	int ret = gnutls_certificate_verify_peers3(tls->tls_session, hostname, &status);
 	if ((ret == GNUTLS_E_SUCCESS) && (status == 0)) {
 		return GNUTLS_E_SUCCESS;
 	}
 
+	const char *addr_str = kr_straddr(session2_get_peer(tls->h.session));
 	if (ret == GNUTLS_E_SUCCESS) {
 		gnutls_datum_t msg;
 		ret = gnutls_certificate_verification_status_print(
-			status, gnutls_certificate_type_get(tls_session), &msg, 0);
+			status, gnutls_certificate_type_get(tls->tls_session), &msg, 0);
 		if (ret == GNUTLS_E_SUCCESS) {
-			kr_log_error(TLSCLIENT, "failed to verify peer certificate: "
-					"%s\n", msg.data);
+			kr_log_error(TLSCLIENT, "failed to verify peer certificate of %s: "
+					"%s\n", addr_str, msg.data);
 			gnutls_free(msg.data);
 		} else {
-			kr_log_error(TLSCLIENT, "failed to verify peer certificate: "
+			kr_log_error(TLSCLIENT, "failed to verify peer certificate of %s: "
 					"unable to print reason: %s (%s)\n",
+					addr_str,
 					gnutls_strerror(ret), gnutls_strerror_name(ret));
 		} /* gnutls_certificate_verification_status_print end */
 	} else {
-		kr_log_error(TLSCLIENT, "failed to verify peer certificate: "
+		kr_log_error(TLSCLIENT, "failed to verify peer certificate of %s: "
 			     "gnutls_certificate_verify_peers3 error: %s (%s)\n",
+			     addr_str,
 			     gnutls_strerror(ret), gnutls_strerror_name(ret));
 	} /* gnutls_certificate_verify_peers3 end */
 	return GNUTLS_E_CERTIFICATE_ERROR;
@@ -849,7 +859,7 @@ static int client_verify_certificate(gnutls_session_t tls_session)
 		/* check hash of the certificate but ignore everything else */
 		return client_verify_pin(cert_list_size, cert_list, tls->client_params);
 	else
-		return client_verify_certchain(tls->tls_session, tls->client_params->hostname);
+		return client_verify_certchain(tls, tls->client_params->hostname);
 }
 
 static int tls_pull_timeout_func(gnutls_transport_ptr_t h, unsigned int ms)
@@ -883,11 +893,25 @@ static int pl_tls_sess_data_deinit(struct pl_tls_sess_data *tls)
 		tls_credentials_release(tls->server_credentials);
 	}
 	wire_buf_deinit(&tls->unwrap_buf);
-	queue_deinit(tls->unwrap_queue); /* TODO: break contexts? */
+
+	while (queue_len(tls->unwrap_queue) > 0) {
+		struct protolayer_iter_ctx *ctx = queue_head(tls->unwrap_queue);
+		protolayer_break(ctx, kr_error(EIO));
+		queue_pop(tls->unwrap_queue);
+	}
+	queue_deinit(tls->unwrap_queue);
+
+	while (queue_len(tls->wrap_queue)) {
+		struct protolayer_iter_ctx *ctx = queue_head(tls->wrap_queue);
+		protolayer_break(ctx, kr_error(EIO));
+		queue_pop(tls->wrap_queue);
+	}
+	queue_deinit(tls->wrap_queue);
+
 	return kr_ok();
 }
 
-static int pl_tls_sess_server_init(struct protolayer_manager *manager,
+static int pl_tls_sess_server_init(struct session2 *session,
                                    struct pl_tls_sess_data *tls)
 {
 	if (kr_fails_assert(the_worker && the_engine))
@@ -965,7 +989,7 @@ static int pl_tls_sess_server_init(struct protolayer_manager *manager,
 					  tls->tls_session);
 	}
 
-	const gnutls_datum_t *alpn = &tls_grp_alpn[manager->grp];
+	const gnutls_datum_t *alpn = &tls_grp_alpn[session->proto];
 	if (alpn->size) { /* ALPN is a non-empty string */
 		flags = 0;
 #if GNUTLS_VERSION_NUMBER >= 0x030500
@@ -985,7 +1009,7 @@ static int pl_tls_sess_server_init(struct protolayer_manager *manager,
 	return kr_ok();
 }
 
-static int pl_tls_sess_client_init(struct protolayer_manager *manager,
+static int pl_tls_sess_client_init(struct session2 *session,
                                    struct pl_tls_sess_data *tls,
                                    tls_client_param_t *param)
 {
@@ -1040,21 +1064,21 @@ static int pl_tls_sess_client_init(struct protolayer_manager *manager,
 	return kr_ok();
 }
 
-static int pl_tls_sess_init(struct protolayer_manager *manager,
+static int pl_tls_sess_init(struct session2 *session,
                             void *sess_data,
                             void *param)
 {
 	struct pl_tls_sess_data *tls = sess_data;
-	manager->session->secure = true;
+	session->secure = true;
 	queue_init(tls->unwrap_queue);
 	queue_init(tls->wrap_queue);
-	if (manager->session->outgoing)
-		return pl_tls_sess_client_init(manager, tls, param);
+	if (session->outgoing)
+		return pl_tls_sess_client_init(session, tls, param);
 	else
-		return pl_tls_sess_server_init(manager, tls);
+		return pl_tls_sess_server_init(session, tls);
 }
 
-static int pl_tls_sess_deinit(struct protolayer_manager *manager,
+static int pl_tls_sess_deinit(struct session2 *session,
                               void *sess_data)
 {
 	return pl_tls_sess_data_deinit(sess_data);
@@ -1065,7 +1089,7 @@ static enum protolayer_iter_cb_result pl_tls_unwrap(void *sess_data, void *iter_
 {
 	int brstatus = kr_ok();
 	struct pl_tls_sess_data *tls = sess_data;
-	struct session2 *s = ctx->manager->session;
+	struct session2 *s = ctx->session;
 
 	queue_push(tls->unwrap_queue, ctx);
 
@@ -1158,7 +1182,7 @@ static enum protolayer_iter_cb_result pl_tls_unwrap(void *sess_data, void *iter_
 	struct protolayer_iter_ctx *ctx_head = queue_head(tls->unwrap_queue);
 	if (!kr_fails_assert(ctx == ctx_head))
 		queue_pop(tls->unwrap_queue);
-	ctx->payload = protolayer_wire_buf(&tls->unwrap_buf, false);
+	ctx->payload = protolayer_payload_wire_buf(&tls->unwrap_buf, false);
 	return protolayer_continue(ctx);
 
 exit_break:
@@ -1172,7 +1196,10 @@ static ssize_t pl_tls_submit(gnutls_session_t tls_session,
                              struct protolayer_payload payload)
 {
 	if (payload.type == PROTOLAYER_PAYLOAD_WIRE_BUF)
-		payload = protolayer_as_buffer(&payload);
+		payload = protolayer_payload_as_buffer(&payload);
+
+	// TODO: the handling of positive gnutls_record_send() is weird/confusing,
+	// but it seems caught later when checking gnutls_record_uncork()
 
 	if (payload.type == PROTOLAYER_PAYLOAD_BUFFER) {
 		ssize_t count = gnutls_record_send(tls_session,
@@ -1276,9 +1303,8 @@ static enum protolayer_event_cb_result pl_tls_client_connect_start(
 
 static enum protolayer_event_cb_result pl_tls_event_unwrap(
 		enum protolayer_event_type event, void **baton,
-		struct protolayer_manager *manager, void *sess_data)
+		struct session2 *s, void *sess_data)
 {
-	struct session2 *s = manager->session;
 	struct pl_tls_sess_data *tls = sess_data;
 
 	if (event == PROTOLAYER_EVENT_CLOSE) {
@@ -1306,7 +1332,7 @@ static enum protolayer_event_cb_result pl_tls_event_unwrap(
 
 static enum protolayer_event_cb_result pl_tls_event_wrap(
 		enum protolayer_event_type event, void **baton,
-		struct protolayer_manager *manager, void *sess_data)
+		struct session2 *session, void *sess_data)
 {
 	if (event == PROTOLAYER_EVENT_STATS_SEND_ERR) {
 		the_worker->stats.err_tls += 1;
@@ -1319,16 +1345,17 @@ static enum protolayer_event_cb_result pl_tls_event_wrap(
 	return PROTOLAYER_EVENT_PROPAGATE;
 }
 
-static void pl_tls_request_init(struct protolayer_manager *manager,
+static void pl_tls_request_init(struct session2 *session,
                                 struct kr_request *req,
                                 void *sess_data)
 {
 	req->qsource.comm_flags.tls = true;
 }
 
-void tls_protolayers_init(void)
+__attribute__((constructor))
+static void tls_protolayers_init(void)
 {
-	protolayer_globals[PROTOLAYER_PROTOCOL_TLS] = (struct protolayer_globals){
+	protolayer_globals[PROTOLAYER_TYPE_TLS] = (struct protolayer_globals){
 		.sess_size = sizeof(struct pl_tls_sess_data),
 		.sess_deinit = pl_tls_sess_deinit,
 		.wire_buf_overhead = TLS_CHUNK_SIZE,
