@@ -30,7 +30,7 @@
  * So it takes 2 RTT.
  * As we use session tickets, there are additional messages, add one RTT mode.
  */
-#define TLS_MAX_HANDSHAKE_TIME (KR_CONN_RTT_MAX * 3)
+#define TLS_MAX_HANDSHAKE_TIME (KR_CONN_RTT_MAX * (uint64_t)3)
 
 /** Transport session (opaque). */
 struct session2;
@@ -103,9 +103,6 @@ int tls_credentials_release(struct tls_credentials *tls_credentials);
 /*! Generate new ephemeral TLS credentials. */
 struct tls_credentials * tls_get_ephemeral_credentials(void);
 
-/*! Initializes the protocol layers managed by tls. */
-void tls_protolayers_init(void);
-
 /* Session tickets, server side.  Implementation in ./tls_session_ticket-srv.c */
 
 /*! Opaque struct used by tls_session_ticket_* functions. */
