@@ -128,7 +128,7 @@ struct worker_ctx *the_worker = NULL;
 /*! @internal Create a UDP/TCP handle for an outgoing AF_INET* connection.
  *  socktype is SOCK_* */
 static struct session2 *ioreq_spawn(int socktype, sa_family_t family,
-                                    enum protolayer_grp grp,
+                                    enum kr_proto grp,
                                     struct protolayer_data_param *layer_param,
                                     size_t layer_param_count)
 {
@@ -194,7 +194,7 @@ static inline struct mempool *pool_borrow(void)
 {
 	/* The implementation used to have extra caching layer,
 	 * but it didn't work well.  Now it's very simple. */
-	return mp_new(16 * 1024);
+	return mp_new((size_t)16 * 1024);
 }
 /** Return a mempool. */
 static inline void pool_release(struct mempool *mp)
@@ -227,25 +227,16 @@ static uint8_t *alloc_wire_cb(struct kr_request *req, uint16_t *maxlen)
 	xdp_handle_data_t *xhd = handle->data;
 	knot_xdp_msg_t out;
 	bool ipv6 = ctx->source.comm_addr.ip.sa_family == AF_INET6;
-	int ret = knot_xdp_send_alloc(xhd->socket,
-			#if KNOT_VERSION_HEX >= 0x030100
-					ipv6 ? KNOT_XDP_MSG_IPV6 : 0, &out);
-			#else
-					ipv6, &out, NULL);
-			#endif
+	int ret = knot_xdp_send_alloc(xhd->socket, ipv6 ? KNOT_XDP_MSG_IPV6 : 0, &out);
 	if (ret != KNOT_EOK) {
 		kr_assert(ret == KNOT_ENOMEM);
 		*maxlen = 0;
 		return NULL;
 	}
 	*maxlen = MIN(*maxlen, out.payload.iov_len);
-#if KNOT_VERSION_HEX < 0x030100
-	/* It's most convenient to fill the MAC addresses at this point. */
-	memcpy(out.eth_from, &ctx->source.eth_from, 6);
-	memcpy(out.eth_to,   &ctx->source.eth_to, 6);
-#endif
 	return out.payload.iov_base;
 }
+
 static void free_wire(const struct request_ctx *ctx)
 {
 	if (kr_fails_assert(ctx->req.alloc_wire_cb == alloc_wire_cb))
@@ -267,12 +258,8 @@ static void free_wire(const struct request_ctx *ctx)
 	out.payload.iov_base = ans->wire;
 	out.payload.iov_len = 0;
 	uint32_t sent = 0;
-#if KNOT_VERSION_HEX >= 0x030100
 	int ret = 0;
 	knot_xdp_send_free(xhd->socket, &out, 1);
-#else
-	int ret = knot_xdp_send(xhd->socket, &out, 1, &sent);
-#endif
 	kr_assert(ret == KNOT_EOK && sent == 0);
 	kr_log_debug(XDP, "freed unsent buffer, ret = %d\n", ret);
 }
@@ -598,7 +585,7 @@ int qr_task_on_send(struct qr_task *task, struct session2 *s, int status)
 						"=> disconnected from '%s': %s\n",
 						peer_str, uv_strerror(status));
 			}
-			worker_end_tcp(s);
+			session2_force_close(s);
 			return status;
 		}
 
@@ -633,9 +620,6 @@ static int qr_task_send(struct qr_task *task, struct session2 *session,
 
 	int ret = 0;
 
-	if (comm == NULL)
-		comm = &session->comm;
-
 	if (pkt == NULL)
 		pkt = worker_task_get_pktbuf(task);
 
@@ -662,7 +646,7 @@ static int qr_task_send(struct qr_task *task, struct session2 *session,
 
 	/* Pending '_finished' callback on current task */
 	qr_task_ref(task);
-	struct protolayer_payload payload = protolayer_buffer(
+	struct protolayer_payload payload = protolayer_payload_buffer(
 			(char *)pkt->wire, pkt->size, false);
 	payload.ttl = packet_ttl(pkt);
 	ret = session2_wrap(session, payload, comm, qr_task_wrap_finished, task);
@@ -832,7 +816,7 @@ static int transmit(struct qr_task *task)
 		return ret;
 
 	struct session2 *session = ioreq_spawn(SOCK_DGRAM, choice->sin6_family,
-			PROTOLAYER_GRP_DOUDP, NULL, 0);
+			KR_PROTO_UDP53, NULL, 0);
 	if (!session)
 		return kr_error(EINVAL);
 
@@ -1088,14 +1072,14 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 	bool has_tls = tls_entry;
 	if (has_tls) {
 		struct protolayer_data_param param = {
-			.protocol = PROTOLAYER_PROTOCOL_TLS,
+			.protocol = PROTOLAYER_TYPE_TLS,
 			.param = tls_entry
 		};
 		session = ioreq_spawn(SOCK_STREAM, addr->sa_family,
-				PROTOLAYER_GRP_DOTLS, &param, 1);
+				KR_PROTO_DOT, &param, 1);
 	} else {
 		session = ioreq_spawn(SOCK_STREAM, addr->sa_family,
-				PROTOLAYER_GRP_DOTCP, NULL, 0);
+				KR_PROTO_TCP53, NULL, 0);
 	}
 	if (!session) {
 		free(conn);
@@ -1303,7 +1287,7 @@ static int qr_task_step(struct qr_task *task,
 
 static int worker_submit(struct session2 *session, struct comm_info *comm, knot_pkt_t *pkt)
 {
-	if (!session || !pkt)
+	if (!session || !pkt || session->closing)
 		return kr_error(EINVAL);
 
 	const bool is_query = pkt->size > KNOT_WIRE_OFFSET_FLAGS1
@@ -1485,16 +1469,6 @@ static struct session2* worker_find_tcp_waiting(const struct sockaddr* addr)
 	return trie_find_tcp_session(the_worker->tcp_waiting, addr);
 }
 
-int worker_end_tcp(struct session2 *session)
-{
-	if (!session)
-		return kr_error(EINVAL);
-
-	session2_timer_stop(session);
-	session2_force_close(session);
-	return kr_ok();
-}
-
 knot_pkt_t *worker_resolve_mk_pkt_dname(knot_dname_t *qname, uint16_t qtype, uint16_t qclass,
 				   const struct kr_qflags *options)
 {
@@ -1710,12 +1684,11 @@ static inline knot_pkt_t *produce_packet(uint8_t *buf, size_t buf_len)
 
 static enum protolayer_event_cb_result pl_dns_dgram_event_unwrap(
 		enum protolayer_event_type event, void **baton,
-		struct protolayer_manager *manager, void *sess_data)
+		struct session2 *session, void *sess_data)
 {
 	if (event != PROTOLAYER_EVENT_GENERAL_TIMEOUT)
 		return PROTOLAYER_EVENT_PROPAGATE;
 
-	struct session2 *session = manager->session;
 	if (session2_tasklist_get_len(session) != 1 ||
 			!session2_waitinglist_is_empty(session))
 		return PROTOLAYER_EVENT_PROPAGATE;
@@ -1753,7 +1726,7 @@ static size_t pl_dns_dgram_wire_buf_overhead(bool outgoing)
 static enum protolayer_iter_cb_result pl_dns_dgram_unwrap(
 		void *sess_data, void *iter_data, struct protolayer_iter_ctx *ctx)
 {
-	struct session2 *session = ctx->manager->session;
+	struct session2 *session = ctx->session;
 
 	if (ctx->payload.type == PROTOLAYER_PAYLOAD_IOVEC) {
 		int ret = kr_ok();
@@ -1772,7 +1745,7 @@ static enum protolayer_iter_cb_result pl_dns_dgram_unwrap(
 				break;
 			}
 
-			ret = worker_submit(session, &ctx->comm, pkt);
+			ret = worker_submit(session, ctx->comm, pkt);
 			if (ret)
 				break;
 		}
@@ -1790,7 +1763,7 @@ static enum protolayer_iter_cb_result pl_dns_dgram_unwrap(
 		if (!pkt)
 			return protolayer_break(ctx, KNOT_EMALF);
 
-		int ret = worker_submit(session, &ctx->comm, pkt);
+		int ret = worker_submit(session, ctx->comm, pkt);
 		mp_flush(the_worker->pkt_pool.ctx);
 		return protolayer_break(ctx, ret);
 	} else if (ctx->payload.type == PROTOLAYER_PAYLOAD_WIRE_BUF) {
@@ -1806,7 +1779,7 @@ static enum protolayer_iter_cb_result pl_dns_dgram_unwrap(
 		if (!pkt)
 			return protolayer_break(ctx, KNOT_EMALF);
 
-		int ret = worker_submit(session, &ctx->comm, pkt);
+		int ret = worker_submit(session, ctx->comm, pkt);
 		wire_buf_reset(ctx->payload.wire_buf);
 		mp_flush(the_worker->pkt_pool.ctx);
 		return protolayer_break(ctx, ret);
@@ -1823,41 +1796,24 @@ struct pl_dns_stream_sess_data {
 	bool connected : 1; /**< True: The stream is connected */
 };
 
-struct pl_dns_stream_iter_data {
-	struct protolayer_data h;
-	struct {
-		knot_mm_t *pool;
-		void *mem;
-	} sent;
-};
-
-static int pl_dns_stream_sess_init(struct protolayer_manager *manager,
-                                         void *sess_data, void *param)
+static int pl_dns_stream_sess_init(struct session2 *session,
+                                   void *sess_data, void *param)
 {
 	/* _UNSIZED_STREAM and _MULTI_STREAM - don't forget to split if needed
 	 * at some point */
-	manager->session->stream = true;
+	session->stream = true;
 	return kr_ok();
 }
 
-static int pl_dns_single_stream_sess_init(struct protolayer_manager *manager,
+static int pl_dns_single_stream_sess_init(struct session2 *session,
                                           void *sess_data, void *param)
 {
-	manager->session->stream = true;
+	session->stream = true;
 	struct pl_dns_stream_sess_data *stream = sess_data;
 	stream->single = true;
 	return kr_ok();
 }
 
-static int pl_dns_stream_iter_deinit(struct protolayer_manager *manager,
-                                     struct protolayer_iter_ctx *ctx,
-                                     void *iter_data)
-{
-	struct pl_dns_stream_iter_data *stream = iter_data;
-	mm_free(stream->sent.pool, stream->sent.mem);
-	return kr_ok();
-}
-
 static enum protolayer_event_cb_result pl_dns_stream_resolution_timeout(
 		struct session2 *s)
 {
@@ -2046,9 +2002,8 @@ static enum protolayer_event_cb_result pl_dns_stream_disconnected(
 
 static enum protolayer_event_cb_result pl_dns_stream_event_unwrap(
 		enum protolayer_event_type event, void **baton,
-		struct protolayer_manager *manager, void *sess_data)
+		struct session2 *session, void *sess_data)
 {
-	struct session2 *session = manager->session;
 	if (session->closing)
 		return PROTOLAYER_EVENT_PROPAGATE;
 
@@ -2056,10 +2011,10 @@ static enum protolayer_event_cb_result pl_dns_stream_event_unwrap(
 
 	switch (event) {
 	case PROTOLAYER_EVENT_GENERAL_TIMEOUT:
-		return pl_dns_stream_resolution_timeout(manager->session);
+		return pl_dns_stream_resolution_timeout(session);
 
 	case PROTOLAYER_EVENT_CONNECT_TIMEOUT:
-		return pl_dns_stream_connection_fail(manager->session,
+		return pl_dns_stream_connection_fail(session,
 				KR_SELECTION_TCP_CONNECT_TIMEOUT);
 
 	case PROTOLAYER_EVENT_CONNECT:
@@ -2069,7 +2024,7 @@ static enum protolayer_event_cb_result pl_dns_stream_event_unwrap(
 		enum kr_selection_error err = (*baton)
 			? *(enum kr_selection_error *)baton
 			: KR_SELECTION_TCP_CONNECT_FAILED;
-		return pl_dns_stream_connection_fail(manager->session, err);
+		return pl_dns_stream_connection_fail(session, err);
 
 	case PROTOLAYER_EVENT_DISCONNECT:
 	case PROTOLAYER_EVENT_CLOSE:
@@ -2171,7 +2126,7 @@ static enum protolayer_iter_cb_result pl_dns_stream_unwrap(
 	}
 
 	int status = kr_ok();
-	struct session2 *session = ctx->manager->session;
+	struct session2 *session = ctx->session;
 	struct pl_dns_stream_sess_data *stream_sess = sess_data;
 	struct wire_buf *wb = ctx->payload.wire_buf;
 
@@ -2192,14 +2147,14 @@ static enum protolayer_iter_cb_result pl_dns_stream_unwrap(
 		if (stream_sess->single && stream_sess->produced) {
 			if (kr_log_is_debug(WORKER, NULL)) {
 				kr_log_debug(WORKER, "Unexpected extra data from %s\n",
-						kr_straddr(ctx->comm.src_addr));
+						kr_straddr(ctx->comm->src_addr));
 			}
 			status = KNOT_EMALF;
 			goto exit;
 		}
 
 		stream_sess->produced = true;
-		int ret = worker_submit(session, &ctx->comm, pkt);
+		int ret = worker_submit(session, ctx->comm, pkt);
 
 		/* Errors from worker_submit() are intentionally *not* handled
 		 * in order to ensure the entire wire buffer is processed. */
@@ -2223,7 +2178,7 @@ exit:
 	wire_buf_movestart(wb);
 	mp_flush(the_worker->pkt_pool.ctx);
 	if (status < 0)
-		worker_end_tcp(session);
+		session2_force_close(session);
 	return protolayer_break(ctx, status);
 }
 
@@ -2235,18 +2190,12 @@ struct sized_iovs {
 static enum protolayer_iter_cb_result pl_dns_stream_wrap(
 		void *sess_data, void *iter_data, struct protolayer_iter_ctx *ctx)
 {
-	struct pl_dns_stream_iter_data *stream = iter_data;
-	struct session2 *s = ctx->manager->session;
-
-	if (kr_fails_assert(!stream->sent.mem))
-		return protolayer_break(ctx, kr_error(EINVAL));
-
 	if (ctx->payload.type == PROTOLAYER_PAYLOAD_BUFFER) {
 		if (kr_fails_assert(ctx->payload.buffer.len <= UINT16_MAX))
 			return protolayer_break(ctx, kr_error(EMSGSIZE));
 
 		const int iovcnt = 2;
-		struct sized_iovs *siov = mm_alloc(&s->pool,
+		struct sized_iovs *siov = mm_alloc(&ctx->pool,
 				sizeof(*siov) + iovcnt * sizeof(struct iovec));
 		kr_require(siov);
 		knot_wire_write_u16(siov->nlen, ctx->payload.buffer.len);
@@ -2259,14 +2208,11 @@ static enum protolayer_iter_cb_result pl_dns_stream_wrap(
 			.iov_len = ctx->payload.buffer.len
 		};
 
-		stream->sent.mem = siov;
-		stream->sent.pool = &s->pool;
-
-		ctx->payload = protolayer_iovec(siov->iovs, iovcnt, false);
+		ctx->payload = protolayer_payload_iovec(siov->iovs, iovcnt, false);
 		return protolayer_continue(ctx);
 	} else if (ctx->payload.type == PROTOLAYER_PAYLOAD_IOVEC) {
 		const int iovcnt = 1 + ctx->payload.iovec.cnt;
-		struct sized_iovs *siov = mm_alloc(&s->pool,
+		struct sized_iovs *siov = mm_alloc(&ctx->pool,
 				sizeof(*siov) + iovcnt * sizeof(struct iovec));
 		kr_require(siov);
 
@@ -2285,10 +2231,7 @@ static enum protolayer_iter_cb_result pl_dns_stream_wrap(
 			.iov_len = sizeof(siov->nlen)
 		};
 
-		stream->sent.mem = siov;
-		stream->sent.pool = &s->pool;
-
-		ctx->payload = protolayer_iovec(siov->iovs, iovcnt, false);
+		ctx->payload = protolayer_payload_iovec(siov->iovs, iovcnt, false);
 		return protolayer_continue(ctx);
 	} else {
 		kr_assert(false && "Invalid payload");
@@ -2296,27 +2239,23 @@ static enum protolayer_iter_cb_result pl_dns_stream_wrap(
 	}
 }
 
-static void pl_dns_stream_request_init(struct protolayer_manager *manager,
+static void pl_dns_stream_request_init(struct session2 *session,
                                        struct kr_request *req,
                                        void *sess_data)
 {
 	req->qsource.comm_flags.tcp = true;
 }
 
-int worker_init(void)
+__attribute__((constructor))
+static void worker_protolayers_init(void)
 {
-	if (kr_fails_assert(the_worker == NULL))
-		return kr_error(EINVAL);
-	kr_bindings_register(the_engine->L); // TODO move
-
-	/* DNS protocol layers */
-	protolayer_globals[PROTOLAYER_PROTOCOL_DNS_DGRAM] = (struct protolayer_globals){
+	protolayer_globals[PROTOLAYER_TYPE_DNS_DGRAM] = (struct protolayer_globals){
 		.wire_buf_overhead_cb = pl_dns_dgram_wire_buf_overhead,
 		.wire_buf_max_overhead = KNOT_WIRE_MAX_PKTSIZE,
 		.unwrap = pl_dns_dgram_unwrap,
 		.event_unwrap = pl_dns_dgram_event_unwrap
 	};
-	protolayer_globals[PROTOLAYER_PROTOCOL_DNS_UNSIZED_STREAM] = (struct protolayer_globals){
+	protolayer_globals[PROTOLAYER_TYPE_DNS_UNSIZED_STREAM] = (struct protolayer_globals){
 		.sess_size = sizeof(struct pl_dns_stream_sess_data),
 		.wire_buf_overhead = KNOT_WIRE_MAX_PKTSIZE,
 		.sess_init = pl_dns_stream_sess_init,
@@ -2326,19 +2265,24 @@ int worker_init(void)
 	};
 	const struct protolayer_globals stream_common = {
 		.sess_size = sizeof(struct pl_dns_stream_sess_data),
-		.iter_size = sizeof(struct pl_dns_stream_iter_data),
 		.wire_buf_overhead = KNOT_WIRE_MAX_PKTSIZE,
 		.sess_init = NULL, /* replaced in specific layers below */
-		.iter_deinit = pl_dns_stream_iter_deinit,
 		.unwrap = pl_dns_stream_unwrap,
 		.wrap = pl_dns_stream_wrap,
 		.event_unwrap = pl_dns_stream_event_unwrap,
 		.request_init = pl_dns_stream_request_init
 	};
-	protolayer_globals[PROTOLAYER_PROTOCOL_DNS_MULTI_STREAM] = stream_common;
-	protolayer_globals[PROTOLAYER_PROTOCOL_DNS_MULTI_STREAM].sess_init = pl_dns_stream_sess_init;
-	protolayer_globals[PROTOLAYER_PROTOCOL_DNS_SINGLE_STREAM] = stream_common;
-	protolayer_globals[PROTOLAYER_PROTOCOL_DNS_SINGLE_STREAM].sess_init = pl_dns_single_stream_sess_init;
+	protolayer_globals[PROTOLAYER_TYPE_DNS_MULTI_STREAM] = stream_common;
+	protolayer_globals[PROTOLAYER_TYPE_DNS_MULTI_STREAM].sess_init = pl_dns_stream_sess_init;
+	protolayer_globals[PROTOLAYER_TYPE_DNS_SINGLE_STREAM] = stream_common;
+	protolayer_globals[PROTOLAYER_TYPE_DNS_SINGLE_STREAM].sess_init = pl_dns_single_stream_sess_init;
+}
+
+int worker_init(void)
+{
+	if (kr_fails_assert(the_worker == NULL))
+		return kr_error(EINVAL);
+	kr_bindings_register(the_engine->L); // TODO move
 
 	/* Create main worker. */
 	the_worker = &the_worker_value;
