@@ -246,11 +246,7 @@ static int pkt_padding(knot_pkt_t *packet, int32_t padding)
 	if (padding == -1) { /* use the default padding policy from libknot */
 		const size_t block_size = knot_wire_get_qr(packet->wire)
 					? KNOT_EDNS_ALIGNMENT_RESPONSE_DEFAULT
-				#if KNOT_VERSION_HEX < 0x030200
-					: KNOT_EDNS_ALIGNMENT_QUERY_DEFALT;
-				#else
 					: KNOT_EDNS_ALIGNMENT_QUERY_DEFAULT;
-				#endif
 		pad_bytes = knot_edns_alignment_size(packet->size, knot_rrset_size(opt_rr),
 							block_size);
 	}
@@ -715,6 +711,8 @@ int kr_resolve_consume(struct kr_request *request, struct kr_transport **transpo
 	if (transport && !qry->flags.CACHED) {
 		if (!(request->state & KR_STATE_FAIL)) {
 			/* Do not complete NS address resolution on soft-fail. */
+			if (kr_fails_assert(packet->wire))
+				return KR_STATE_FAIL;
 			const int rcode = knot_wire_get_rcode(packet->wire);
 			if (rcode != KNOT_RCODE_SERVFAIL && rcode != KNOT_RCODE_REFUSED) {
 				qry->flags.AWAIT_IPV6 = false;
@@ -748,7 +746,7 @@ int kr_resolve_consume(struct kr_request *request, struct kr_transport **transpo
 	}
 
 	/* Pop query if resolved. */
-	if (request->state == KR_STATE_YIELD) {
+	if (request->state == KR_STATE_YIELD) { // NOLINT(bugprone-branch-clone)
 		return KR_STATE_PRODUCE; /* Requery */
 	} else if (qry->flags.RESOLVED) {
 		kr_rplan_pop(rplan, qry);
@@ -931,6 +929,7 @@ int kr_resolve_finish(struct kr_request *request, int state)
 				knot_wire_clear_ad(wire);
 				knot_wire_clear_aa(wire);
 				knot_wire_set_rcode(wire, KNOT_RCODE_SERVFAIL);
+			default:; // Do nothing
 			}
 		}
 	}
