@@ -17,8 +17,15 @@ def _get_templates_dir() -> str:
 _TEMPLATES_DIR = _get_templates_dir()
 
 
-def _import_main_template() -> Template:
-    path = os.path.join(_TEMPLATES_DIR, "config.lua.j2")
+def _import_kresd_worker_config_template() -> Template:
+    path = os.path.join(_TEMPLATES_DIR, "worker-config.lua.j2")
+    with open(path, "r", encoding="UTF-8") as file:
+        template = file.read()
+    return template_from_str(template)
+
+
+def _import_kresd_policy_config_template() -> Template:
+    path = os.path.join(_TEMPLATES_DIR, "policy-config.lua.j2")
     with open(path, "r", encoding="UTF-8") as file:
         template = file.read()
     return template_from_str(template)
@@ -30,4 +37,7 @@ def template_from_str(template: str) -> Template:
     return env.from_string(template)
 
 
-MAIN_TEMPLATE = _import_main_template()
+WORKER_CONFIG_TEMPLATE = _import_kresd_worker_config_template()
+
+
+POLICY_CONFIG_TEMPLATE = _import_kresd_policy_config_template()
