@@ -188,7 +188,6 @@ class Server:
             }
         )
 
-    @statistics.async_timing_histogram(statistics.MANAGER_REQUEST_RECONFIGURE_LATENCY)
     async def _handler_config_query(self, request: web.Request) -> web.Response:
         """
         Route handler for changing resolver configuration
@@ -238,9 +237,24 @@ class Server:
         res.headers.add("ETag", f'"{structural_etag(new_config)}"')
         return res
 
-    async def _handler_metrics(self, _request: web.Request) -> web.Response:
+    async def _handler_metrics(self, request: web.Request) -> web.Response:
+        raise web.HTTPMovedPermanently("/metrics/json")
+
+    async def _handler_metrics_json(self, _request: web.Request) -> web.Response:
         return web.Response(
             body=await statistics.report_stats(),
+            content_type="application/json",
+            charset="utf8",
+        )
+
+    async def _handler_metrics_prometheus(self, _request: web.Request) -> web.Response:
+
+        metrics_report = await statistics.report_stats(prometheus_format=True)
+        if not metrics_report:
+            raise web.HTTPNotFound()
+
+        return web.Response(
+            body=metrics_report,
             content_type="text/plain",
             charset="utf8",
         )
@@ -330,6 +344,8 @@ class Server:
                 web.get("/schema", self._handler_schema),
                 web.get("/schema/ui", self._handle_view_schema),
                 web.get("/metrics", self._handler_metrics),
+                web.get("/metrics/json", self._handler_metrics_json),
+                web.get("/metrics/prometheus", self._handler_metrics_prometheus),
                 web.post("/cache/clear", self._handler_cache_clear),
             ]
         )
