@@ -1,6 +1,6 @@
 import pytest
 
-from knot_resolver_manager.config_store import ConfigStore, only_on_real_changes
+from knot_resolver_manager.config_store import ConfigStore, only_on_real_changes_update
 from knot_resolver_manager.datamodel.config_schema import KresConfig
 
 
@@ -8,7 +8,7 @@ from knot_resolver_manager.datamodel.config_schema import KresConfig
 async def test_only_once():
     count = 0
 
-    @only_on_real_changes(lambda config: config.logging.level)
+    @only_on_real_changes_update(lambda config: config.logging.level)
     async def change_callback(config: KresConfig) -> None:
         nonlocal count
         count += 1
